/**
 *
 * Initial version of this code (c) 2009-2011 Media Tuners LLC with a full license to Pioneer Corporation.
 *
 * Pioneer Corporation licenses this file to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is
 * distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and limitations under the License.
 *
 */

package net.zypr.gui.windows;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.awt.event.MouseListener;
import java.awt.image.BufferedImage;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.swing.*;

import net.miginfocom.swing.MigLayout;

import net.zypr.api.API;
import net.zypr.api.Weather;
import net.zypr.api.enums.TemperatureScale;
import net.zypr.api.vo.WeatherForecastVO;
import net.zypr.gui.ImageFetcher;
import net.zypr.gui.Settings;
import net.zypr.gui.components.ImagePanel;
import net.zypr.gui.components.Label;
import net.zypr.gui.utils.Converter;
import net.zypr.gui.utils.WebBrowser;
import org.jdesktop.swingx.JXPanel;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WeatherForecastWindow extends ModalWindow
{
    private final Logger _logger = LoggerFactory.getLogger(getClass().getCanonicalName());
    private WeatherForecastVO[] _weatherForecastVOs = null;
    private String _service = null;
    private String _placename = null;
    private ForecastDayPanel forecastdays[];

    public WeatherForecastWindow(String service, WeatherForecastVO[] weatherForecastVOs, String placename)
    {
        super();
        try {
            _service = service;
            _weatherForecastVOs = weatherForecastVOs;
            _placename = placename;
            jbInit();
        } catch (Exception exception) {
            _logger.error("", exception);
        }
    }

    private class ForecastDayPanel extends JXPanel implements MouseListener
    {

        private ImagePanel dayImage = new ImagePanel();
        private ImagePanel nightImage = new ImagePanel();
        private Label dayofweekLabel = new Label();
        private Label dateLabel = new Label();
        private Label hiTempLabel = new Label();
        private Label loTempLabel = new Label();
        private String dayDetailsUri;
        private String nightDetailsUri;

        public ForecastDayPanel()
        {
            MigLayout layout = new MigLayout("flowy, insets 5", "[100]", "[fill]0[fill]");
            this.setLayout(layout);
            this.setBackground(new Color(0, 0, 0, 0));
            this.add(dayofweekLabel, "width 80!");
            this.add(dateLabel, "width 80!");
            this.add(dayImage, "width 80!, height 80!");
            this.add(hiTempLabel, "width 80!");
            this.add(nightImage, "width 80!, height 80!");
            this.add(loTempLabel, "width 80!");
            this.addMouseListener(this);
        }

        public void setDayDetailsUri(String uri)
        {
            dayDetailsUri = uri;
        }

        public void setNightDetailsUri(String uri)
        {
            nightDetailsUri = uri;
        }

        public void setDayImage(BufferedImage day)
        {
            dayImage.setBackground(new Color(0, 0, 0, 0));
            dayImage.setImage(day);
        }

        public void setNightImage(BufferedImage day)
        {
            nightImage.setBackground(new Color(0, 0, 0, 0));
            nightImage.setImage(day);
        }

        public void setDate(Date date)
        {
            SimpleDateFormat simpleDateFormat = new SimpleDateFormat("EEEE");
            dayofweekLabel.setText(simpleDateFormat.format(date));
            dayofweekLabel.setHorizontalAlignment(SwingConstants.CENTER);
            dayofweekLabel.setFont(new Font("Dialog", Font.PLAIN, 14));

            simpleDateFormat = new SimpleDateFormat("MMM dd");
            dateLabel.setText(simpleDateFormat.format(date));
            dateLabel.setHorizontalAlignment(SwingConstants.CENTER);
            dateLabel.setFont(new Font("Dialog", Font.PLAIN, 10));
        }

        public void setTemperature(TemperatureScale scale, long hi, long lo)
        {

            TemperatureScale tempDisplayScale = Settings.getInstance().getDTO().getTemperatureScale();
            long dispHi = (long) Converter.convertTemperatures(scale, tempDisplayScale, hi);
            long dispLo = (long) Converter.convertTemperatures(scale, tempDisplayScale, lo);

            hiTempLabel.setText(dispHi + tempDisplayScale.getSymbol() + " Hi");
            hiTempLabel.setHorizontalAlignment(SwingConstants.CENTER);
            hiTempLabel.setFont(new Font("Dialog", Font.PLAIN, 12));

            loTempLabel.setText(dispLo + tempDisplayScale.getSymbol() + " Lo");
            loTempLabel.setHorizontalAlignment(SwingConstants.CENTER);
            loTempLabel.setFont(new Font("Dialog", Font.PLAIN, 12));
        }

        @Override
        public void mouseClicked(MouseEvent mouseEvent)
        {
            _logger.debug("Mouse clicked {}", mouseEvent.paramString());
            if (mouseEvent.getY() > 130) {
                WebBrowser.openURL(nightDetailsUri);
            } else {
                WebBrowser.openURL(dayDetailsUri);
            }
        }

        @Override
        public void mousePressed(MouseEvent mouseEvent)
        {
        }

        @Override
        public void mouseReleased(MouseEvent mouseEvent)
        {
        }

        @Override
        public void mouseEntered(MouseEvent mouseEvent)
        {
        }

        @Override
        public void mouseExited(MouseEvent mouseEvent)
        {
        }
    }

    private void jbInit() throws Exception
    {
        this.setSize(new Dimension(560, 300));
        setTitle("5 Day Forecast for " + _placename);

        MigLayout layout = new MigLayout("flowx, insets 5");
        _panelContent.setLayout(layout);

        // TODO: run this in a separate thread
        if (_service == null || _service.equalsIgnoreCase(Weather.SERVICE_NAME_DEFAULT))
            _service = API.getInstance().getWeather().getDefaultServiceForCurrent();
        setTitleImage(ImageFetcher.getInstance().getImage(API.getInstance().getWeather().getService(_service).getServiceIconURI()));

        forecastdays = new ForecastDayPanel[5];
        for (int i = 0; i < 5; i++) {
            forecastdays[i] = new ForecastDayPanel();
            forecastdays[i].setDayImage(ImageFetcher.getInstance().getImage(_weatherForecastVOs[2 * i].getIconURI(), 80, 80, true));
            forecastdays[i].setNightImage(ImageFetcher.getInstance().getImage(_weatherForecastVOs[2 * i + 1].getIconURI(), 80, 80, true));
            forecastdays[i].setDate(_weatherForecastVOs[2 * i].getDate());
            forecastdays[i].setTemperature(_weatherForecastVOs[2 * i].getScale(), _weatherForecastVOs[2 * i].getHighTemperature(), _weatherForecastVOs[2 * i].getLowTemperature());
            forecastdays[i].setDayDetailsUri(_weatherForecastVOs[2 * i].getDetailsURI());
            forecastdays[i].setNightDetailsUri(_weatherForecastVOs[2 * i + 1].getDetailsURI());
            _panelContent.add(forecastdays[i]);
        }
    }

    protected void _buttonWindowClose_actionPerformed(ActionEvent actionEvent)
    {
        getParentPanel().discardWindow();
    }
}
